import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Queue;
import java.util.TreeSet;
import java.util.LinkedList;

public class graph {

	public static final int maxDistance = Integer.MAX_VALUE;
	private static int N = 0;

	public static String pairToString(int a, int b) {
		String value = new String("");

		if (a < b) {
			value = a + " " + b;
		} else {
			value = b + " " + a;
		}

		return value;
	}

	public void findShortestPath(int start, int destination,
			ArrayList<City> country, ArrayList<Integer> minDistance,
			ArrayList<Integer> previous, HashMap<String, Integer> B) {

		int n = country.size() + 1;
		int currentCity = start;
		int currentDistance = 0;

		minDistance = new ArrayList<Integer>(n);
		for (int i = 0; i < n; i++) {
			minDistance.set(i, maxDistance);
		}
		minDistance.set(start, 0);

		previous = new ArrayList<Integer>(n);

		TreeSet<Distance> closestCityQueue = new TreeSet<Distance>(
				new DistanceComparator());

		closestCityQueue.add(new Distance(start, 0));

		while (!closestCityQueue.isEmpty()) {
			currentCity = closestCityQueue.first().getCity();
			currentDistance = closestCityQueue.first().getDistance();
			closestCityQueue.remove(closestCityQueue.first());

			ArrayList<Neighbour> Neighbours = country.get(currentCity)
					.getNeighbours();
			for (int i = 0; i < Neighbours.size(); i++) {
				int v = Neighbours.get(i).getCity();
				int weight = Neighbours.get(i).getDistance();
				int distanceThroughV = currentDistance + weight;

				if (distanceThroughV < minDistance.get(v)) {
					closestCityQueue
							.remove(new Distance(v, minDistance.get(v)));

					minDistance.set(v, distanceThroughV);

					previous.set(v, currentCity);

					closestCityQueue.add(new Distance(v, minDistance.get(v)));
				}
			}
		}

		for (int i = start; i < destination; i++) {
			B.put(pairToString(start, i), (Integer) minDistance.get(i));
		}
	}

	public static ArrayList<Integer> getShortestPathTo(int vertex,
			ArrayList<Integer> previous) {
		ArrayList<Integer> path = new ArrayList<Integer>();
		for (; vertex != -1; vertex = previous.get(vertex))
			path.add(vertex);
		return path;
	}

	public static HashMap<String, Integer> loadParameters(
			ArrayList<City> country, HashMap<String, Integer> A,
			HashMap<String, Integer> roads) throws IOException {

		FileReader file;
		file = new FileReader("graph.in");
		BufferedReader input = new BufferedReader(file);

		int N, M, a, b, c, dist;

		String[] firstRow = input.readLine().split(" ");

		N = Integer.parseInt(firstRow[0]);
		M = Integer.parseInt(firstRow[1]);

		country.ensureCapacity(N + 1);

		for (int i = 0; i < N + 1; i++) {
			country.add(new City(i));
		}

		roads = new HashMap<String, Integer>(M + 1);

		for (int i = 0; i < country.size(); i++) {
			country.get(i).setId(i);
		}

		for (int i = 1; i < M + 1; i++) {
			String[] roadRow = input.readLine().split(" ");
			a = Integer.parseInt(roadRow[0]);
			b = Integer.parseInt(roadRow[1]);
			c = Integer.parseInt(roadRow[2]);

			roads.put(pairToString(a, b), i);

			country.get(a).addNeighbour(new Neighbour(b, c));
			country.get(b).addNeighbour(new Neighbour(a, c));
		}

		A.clear();

		for (int i = 0; i < N; i++) {
			String[] traffic = input.readLine().split(" ");
			for (int j = 0; j < N; j++) {
				dist = Integer.parseInt(traffic[j]);
				A.put(pairToString(i + 1, j + 1), dist);
			}
		}

		input.close();

		graph.N = N;
		return roads;
	}

	public static ArrayList<ArrayList<Integer>> generateShortestPathsMatrix(
			ArrayList<City> country, int N,
			ArrayList<ArrayList<Integer>> paths, HashMap<String, Integer> B)
			throws InterruptedException {

		ArrayList<ArrayList<Integer>> minDistance = new ArrayList<ArrayList<Integer>>(
				N);
		for (int i = 0; i < N; i++) {
			minDistance.add(new ArrayList<Integer>());
		}

		paths = new ArrayList<ArrayList<Integer>>();
		for (int i = 0; i < N + 1; i++) {
			paths.add(new ArrayList<Integer>());
		}

		Thread[] threads = new Thread[N + 1];
		Object lock1 = new Object();
		Object lock2 = new Object();
		Object lock3 = new Object();

		for (int i = 0; i < N; i++) {
			if (i < 6) {
				threads[i] = new ShortestPathThread(i + 1, N, country,
						minDistance.get(i), paths.get(i + 1), B, lock1, lock2,
						lock3);
				threads[i].start();
			} else {
				threads[i - 6].join();
				threads[i] = new ShortestPathThread(i + 1, N, country,
						minDistance.get(i), paths.get(i + 1), B, lock1, lock2,
						lock3);
				threads[i].start();
			}
		}

		if(N >= 6) threads[N - 6].join();
		if(N >= 5) threads[N - 5].join();
		if(N >= 4) threads[N - 4].join();
		if(N >= 3) threads[N - 3].join();
		threads[N - 2].join();
		threads[N - 1].join();

		for (int i = 1; i <= N; i++) {
			paths.set(i, ((ShortestPathThread) threads[i - 1]).getPrevious());
		}

		return paths;
	}

	public static boolean checkConnection(int start, int end,
			ArrayList<City> country) {

		ArrayList<Boolean> cities = new ArrayList<Boolean>(country.size() + 1);
		for (int i = 0; i < country.size() + 1; i++) {
			cities.add(false);
		}
		Queue<City> cityQueue = new LinkedList<City>();

		for (int i = 0; i < cities.size(); i++) {
			cities.set(i, false);
		}

		City current = country.get(start);
		cities.set(start, true);
		cityQueue.add(current);

		while (!cityQueue.isEmpty()) {
			current = cityQueue.poll();
			if (current.getId() == end) {
				return true;
			}

			ArrayList<Neighbour> Neighbours = current.getNeighbours();
			for (int i = 0; i < Neighbours.size(); i++) {
				if (!cities.get(Neighbours.get(i).getCity())) {
					cityQueue.add(country.get(Neighbours.get(i).getCity()));
				}
				cities.set(Neighbours.get(i).getCity(), true);
			}
		}

		return false;
	}

	public static void generateCoefficientsMatrix(HashMap<String, Integer> A,
			HashMap<String, Integer> B, TreeSet<Coefficient> C, int N) {

		C.clear();

		for (int i = 0; i < N; i++) {
			for (int j = i; j < N; j++) {
				C.add(new Coefficient(i + 1, j + 1, (double) A
						.get(pairToString(i + 1, j + 1))
						* B.get(pairToString(i + 1, j + 1))));
			}
		}
	}

	public static boolean checkConnections(int start, ArrayList<City> country,
			int[] connected, int N) {

		Queue<City> cityQueue = new LinkedList<City>();

		connected = new int[N + 1];

		City current = country.get(start);
		connected[start] = 1;
		cityQueue.add(current);

		while (!cityQueue.isEmpty()) {
			current = cityQueue.poll();

			ArrayList<Neighbour> Neighbours = current.getNeighbours();
			for (int i = 0; i < Neighbours.size(); i++) {
				if (connected[Neighbours.get(i).getCity()] == 0) {
					cityQueue.add(country.get(Neighbours.get(i).getCity()));
				}
				connected[Neighbours.get(i).getCity()] = 1;
			}
		}

		for (int i = 1; i < N + 1; i++) {
			if (connected[i] == 0) {
				return false;
			}
		}

		return true;
	}

	public static void generateFinalGraph(TreeSet<Coefficient> C,
			ArrayList<ArrayList<Integer>> paths,
			HashMap<String, Integer> roads, int N) throws IOException {

		int edgeCount = 0;

		FileWriter file = new FileWriter("graph.out");
		BufferedWriter output = new BufferedWriter(file);

		int[] counter = new int[N + 1];
		for (int i = 0; i < N + 1; i++) {
			counter[i] = 0;
		}

		ArrayList<City> newGraph = new ArrayList<City>(N + 1);
		for (int i = 0; i < N + 1; i++) {
			newGraph.add(new City(i));
		}

		while (edgeCount < (N - 1)) {
			Coefficient coef = C.first();
			int[] p = coef.getIndexes();
			C.remove(C.first());
			System.out.println(p[0] + " " + p[1]);
			if (counter[p[0]] == 1 && counter[p[1]] == 1) {
				continue;
			}
	
			ArrayList<Integer> v = getShortestPathTo(p[0], paths.get(p[1]));
			int len = v.size();

			if (len == 2) {
				if (counter[v.get(0)] == 0 || counter[v.get(1)] == 0
						|| !checkConnection(v.get(0), v.get(1), newGraph)) {
					if (edgeCount < N - 2) {
						output.write(roads
								.get(pairToString(v.get(0), v.get(1)))
								.toString()
								+ "\n");
					} else {
						output.write(roads
								.get(pairToString(v.get(0), v.get(1)))
								.toString());
					}
					newGraph.get(v.get(0)).addNeighbour(
							new Neighbour(v.get(1), 1));
					newGraph.get(v.get(1)).addNeighbour(
							new Neighbour(v.get(0), 1));
					edgeCount++;
					counter[v.get(0)] += 1;
					counter[v.get(1)] += 1;
				}
			} else {
				for (int i = 0; i < len - 1; i++) {
					if (counter[v.get(i)] == 0
							|| counter[v.get(i + 1)] == 0
							|| !checkConnection(v.get(i), v.get(i + 1),
									newGraph)) {
						if (edgeCount < N - 2) {
							output.write(roads.get(
									pairToString(v.get(i), v.get(i + 1)))
									.toString()
									+ "\n");
						} else {
							output.write(roads.get(
									pairToString(v.get(i), v.get(i + 1)))
									.toString());
						}
						newGraph.get(v.get(i)).addNeighbour(
								new Neighbour(v.get(i + 1), 1));
						newGraph.get(v.get(i + 1)).addNeighbour(
								new Neighbour(v.get(i), 1));
						edgeCount++;
						counter[v.get(i)] += 1;
						counter[v.get(i + 1)] += 1;
					}
				}
			}
		}

		if (edgeCount == N - 1) {
			output.close();
			return;
		}

		int index;
		for (index = 0; index < N; index++) {
			if (counter[index] == 1) {
				break;
			}
		}

		checkConnections(index, newGraph, counter, N);

		for (int i = 1; i < N + 1; i++) {
			if (counter[i] == 0) {

				ArrayList<Neighbour> Neighbours = newGraph.get(i)
						.getNeighbours();
				for (int j = 0; j < Neighbours.size(); j++) {
					counter[Neighbours.get(j).getCity()] = 1;
				}

				ArrayList<Integer> v = getShortestPathTo(index, paths.get(i));
				int len = v.size();

				if (len == 2) {
					if (counter[v.get(0)] == 0 || counter[v.get(1)] == 0
							|| !checkConnection(v.get(0), v.get(1), newGraph)) {
						if (edgeCount < N - 2) {
							output.write(roads.get(
									pairToString(v.get(0), v.get(1)))
									.toString()
									+ "\n");
						} else {
							output.write(roads.get(
									pairToString(v.get(0), v.get(1)))
									.toString());
						}
						newGraph.get(v.get(0)).addNeighbour(
								new Neighbour(v.get(1), 1));
						newGraph.get(v.get(1)).addNeighbour(
								new Neighbour(v.get(0), 1));
						edgeCount++;
						counter[v.get(0)] += 1;
						counter[v.get(1)] += 1;
					}
				} else {
					for (int j = 0; j < len - 1; j++) {
						if (counter[v.get(j)] == 0
								|| counter[v.get(j + 1)] == 0
								|| !checkConnection(v.get(j), v.get(j + 1),
										newGraph)) {
							if (edgeCount < N - 2) {
								output.write(roads.get(
										pairToString(v.get(j), v.get(j + 1)))
										.toString()
										+ "\n");
							} else {
								output.write(roads.get(
										pairToString(v.get(j), v.get(j + 1)))
										.toString());
							}
							newGraph.get(v.get(j)).addNeighbour(
									new Neighbour(v.get(j + 1), 1));
							newGraph.get(v.get(j + 1)).addNeighbour(
									new Neighbour(v.get(j), 1));
							edgeCount++;
							counter[v.get(j)] += 1;
							counter[v.get(j + 1)] += 1;
						}
					}
				}
			}
		}

		output.close();
	}

	public static void main(String[] args) {

		ArrayList<City> graph = new ArrayList<City>();

		HashMap<String, Integer> B = new HashMap<String, Integer>();
		ArrayList<ArrayList<Integer>> paths = new ArrayList<ArrayList<Integer>>();
		HashMap<String, Integer> A = new HashMap<String, Integer>();
		TreeSet<Coefficient> C = new TreeSet<Coefficient>(
				new CoefficientComparator());

		HashMap<String, Integer> roads = new HashMap<String, Integer>();

		try {
			roads = loadParameters(graph, A, roads);

		} catch (IOException e) {
			e.printStackTrace();
		}

		try {
			paths = generateShortestPathsMatrix(graph, N, paths, B);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}

		generateCoefficientsMatrix(A, B, C, N);

		try {
			generateFinalGraph(C, paths, roads, N);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private static class City {
		private int id;
		ArrayList<Neighbour> neighbours = new ArrayList<Neighbour>();

		City() {

		}

		City(int id) {
			this.id = id;
		}

		void setId(int id) {
			this.id = id;
		}

		void setNeighbours(ArrayList<Neighbour> neighbours) {
			this.neighbours = new ArrayList<Neighbour>(neighbours);
		}

		void addNeighbour(Neighbour n) {
			this.neighbours.add(n);
		}

		int getId() {
			return this.id;
		}

		ArrayList<Neighbour> getNeighbours() {
			return this.neighbours;
		}

	}

	private static class ShortestPathThread extends Thread {

		private int start;
		private int destination;
		private ArrayList<City> country;
		private ArrayList<Integer> minDistance;
		private ArrayList<Integer> previous;
		private HashMap<String, Integer> B;
		private Object lock1, lock2, lock3;

		public ShortestPathThread() {

		}

		public ShortestPathThread(int start, int destination,
				ArrayList<City> country, ArrayList<Integer> minDistance,
				ArrayList<Integer> previous, HashMap<String, Integer> B, Object l1,
				Object l2, Object l3) {
			this.start = start;
			this.country = country;
			this.destination = destination;
			this.country = country;
			this.minDistance = minDistance;
			this.previous = previous;
			this.B = B;
			this.lock1 = l1;
			this.lock2 = l2;
			this.lock3 = l3;
		}

		@Override
		public void run() {

			int n = country.size() + 1;
			int currentCity = start;
			int currentDistance = 0;

			lock1.toString();

			synchronized (lock1) {
				minDistance = new ArrayList<Integer>(n);
				for (int i = 0; i < n; i++) {
					minDistance.add(Integer.MAX_VALUE);
				}
				minDistance.set(start, 0);
			}

			synchronized (lock2) {
				previous = new ArrayList<Integer>(n);
				for (int i = 0; i < n; i++) {
					previous.add(-1);
				}
			}

			TreeSet<Distance> closestCityQueue = new TreeSet<Distance>(
					new DistanceComparator());

			closestCityQueue.add(new Distance(start, 0));

			while (!closestCityQueue.isEmpty()) {
				currentCity = closestCityQueue.first().getCity();
				currentDistance = closestCityQueue.first().getDistance();
				closestCityQueue.remove(closestCityQueue.first());

				ArrayList<Neighbour> Neighbours = country.get(currentCity)
						.getNeighbours();
				for (int i = 0; i < Neighbours.size(); i++) {
					int v = Neighbours.get(i).getCity();
					int weight = Neighbours.get(i).getDistance();
					int distanceThroughV = currentDistance + weight;

					if (distanceThroughV < minDistance.get(v)) {
						closestCityQueue
								.remove(new Distance(v, minDistance.get(v)));

						synchronized (lock1) {
							minDistance.set(v, distanceThroughV);
						}

						synchronized (lock2) {
							previous.set(v, currentCity);
						}

						closestCityQueue.add(new Distance(v, minDistance.get(v)));
					}
				}
			}

			synchronized (lock3) {
				for (int i = start; i <= destination; i++) {
					B.put(graph.pairToString(start, i),
							(Integer) minDistance.get(i));
				}
			}
		}

		public ArrayList<Integer> getPrevious() {
			return this.previous;
		}

	}

	private static class Coefficient {
		private int indexOne;
		private int indexTwo;
		private double coef;

		Coefficient() {

		}

		Coefficient(int i1, int i2, double c) {
			this.indexOne = i1;
			this.indexTwo = i2;
			this.coef = c;
		}

		public int[] getIndexes() {
			int[] indexes = new int[2];
			indexes[0] = this.indexOne;
			indexes[1] = this.indexTwo;
			return indexes;
		}

		public double getCoefficient() {
			return this.coef;
		}

	}

	private static class CoefficientComparator implements Comparator<Coefficient> {

		@Override
		public int compare(Coefficient arg0, Coefficient arg1) {
			return ((Double) arg1.getCoefficient()).compareTo((Double) arg0
					.getCoefficient());
		}

	}

	private static class Distance {

		private int city;
		private int distance;

		public Distance() {

		}

		public Distance(int city, int distance) {
			this.city = city;
			this.distance = distance;
		}

		public void setCity(int city) {
			this.city = city;
		}

		public void setDistance(int distance) {
			this.distance = distance;
		}

		public int getCity() {
			return this.city;
		}

		public int getDistance() {
			return this.distance;
		}

		@Override
		public boolean equals(Object o) {
			if (this.distance == ((Distance) o).getDistance()
					&& this.city == ((Distance) o).getCity()) {
				return true;
			} else {
				return false;
			}
		}
	}

	private static class DistanceComparator implements Comparator<Distance> {

		@Override
		public int compare(Distance o1, Distance o2) {
			if (o1.getDistance() > o2.getDistance()) {
				return 1;
			} else {
				if (o1.getDistance() < o2.getDistance()) {
					return -1;
				} else {
					return 0;
				}
			}
		}
	}

	private static class Neighbour {
		private int city;
		private int distance;

		Neighbour() {

		}

		Neighbour(int city, int distance) {
			this.city = city;
			this.distance = distance;
		}

		void setCity(int city) {
			this.city = city;
		}

		void setDistance(int distance) {
			this.distance = distance;
		}

		int getCity() {
			return this.city;
		}

		int getDistance() {
			return this.distance;
		}

	}


}
